/***************************************************************************\
  "icm.h" - Intermediate Code Manager (Phase D)
  Revision 14; 10 Nov 2002 - 21 Feb 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

#ifndef ICM_H
#define ICM_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdlib.h>

// Local //
#include "global.h"
#include "error.h"
#include "table.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //
typedef struct
{
  // statement as it appears in input code //
  TokenCode code;
  Symbol *sym;

  // where this statement occured in the input file //
  char *file;
  int line;

} Statement;

Statement *statement_new (TokenCode aCode, Symbol *aSym);
void statement_del (Statement *statement);

Statement** gvProgram; // vector of statements
int gxLast;            // index of last statement in the gvProgram vector
int gxCurr;            // index of current statement in the gvProgram vector

void icm_init ();
void icm_deinit ();
int icm_put (Statement *statement);
void icm_putline ();
void icm_reset ();
void icm_goto (int x);
Statement *icm_getnext ();
Statement *icm_get (int x);

// Internal //
bool mReset;

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// statement_new
Statement *statement_new (TokenCode aCode, Symbol *aSym)
{
  Statement *temp;

  temp = (Statement *) KALLOC(sizeof (Statement));
  temp->code = aCode;
  temp->sym = aSym;

  if (mReset == false)
  {
    if ((gxCurr + 1) > GLOBAL_ICM_MAX_INSTRUCTIONS)
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }
    else
    {
      temp->line = gnCurrentLine;
      temp->file = k_clone (gsCurrentFilename);
    }
  }
  else
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    erre (EFF_TERMINATE);
  }

  return temp;
}

//// statement_del
void statement_del (Statement *statement)
{
  if (statement)
  {
    k_del (statement->file);

    // NOTE: Symbols are not deleted here; they are deleted in   //
    //       scope_deinit() by way of scope_add_symbol_to_list() //

    KFREE(statement);
  }
}

//// icm_init
void icm_init ()
{
  int i;

  gvProgram = (Statement **) KALLOC(sizeof (Statement *) * GLOBAL_ICM_MAX_INSTRUCTIONS);
  for (i = 0; i < GLOBAL_ICM_MAX_INSTRUCTIONS; i++)
  {
    gvProgram[i] = NULL;
  }

  gxLast = 0;
  gxCurr = 0;
  mReset = false;
}

//// icm_deinit
void icm_deinit ()
{
  int i;

  for (i = 0; i < GLOBAL_ICM_MAX_INSTRUCTIONS; i++)
  {
    if (gvProgram[i]) statement_del (gvProgram[i]);
  }
  KFREE(gvProgram);
}

//// icm_put
int icm_put (Statement *statement)
{
  if (mReset == false)
  {
    if ((gxCurr + 1) > GLOBAL_ICM_MAX_INSTRUCTIONS)
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    gvProgram[gxCurr] = statement;
    gxCurr++;
    gxLast++;
  }
  else
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    erre (EFF_TERMINATE);
  }

  return (gxCurr - 1);
}

//// icm_reset
void icm_reset ()
{
  gxCurr = 0;
  mReset = true;
}

//// icm_goto
void icm_goto (int x)
{
  if (mReset == true)
  {
    if ((x > gxLast) || (x > GLOBAL_ICM_MAX_INSTRUCTIONS))
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    gxCurr = x;
  }
  else
  {
    err (true, EI_GOTO_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
  }
}

//// icm_getnext
Statement *icm_getnext ()
{
  if (mReset == true)
  {
    if ((gxCurr + 1) > gxLast)
    {
      return NULL; // there are no more statements, so return NULL
    }
    else
    {
      Statement *temp = gvProgram[gxCurr++];

      gnCurrentLine = temp->line;
      k_copy (gsCurrentFilename, temp->file);

      return temp;
    }
  }
  else
  {
    err (true, EI_READ_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
  }

  return NULL;
}

//// icm_get
Statement *icm_get (int x)
{
  if (mReset == true)
  {
    if ((x > gxLast) || (x > GLOBAL_ICM_MAX_INSTRUCTIONS))
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    return gvProgram[x];
  }
  else
  {
    err (true, EI_READ_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
    return NULL;
  }
}

#endif
