/***************************************************************************\
  "lexer.h" - Lexical Analyzer (Phase B)
  Revision 17; 10 Nov 2002 - 21 Feb 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht

  References:
    MCD 2.1.5; 2.1.5.1
\***************************************************************************/

#ifndef LEXER_H
#define LEXER_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

#define LEXER_DIGITS "0123456789"
#define LEXER_NUMBER_ELEMENTS "0123456789."
#define LEXER_LETTERS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_"
#define LEXER_SPECIALS "+-*/\\=(),<>$.[]\":;{}&#^"
#define LEXER_MIN_WORD_LEN 2
#define LEXER_MAX_WORD_LEN 7

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdlib.h>
#include <math.h>

// Local //
#include "inputbuffer.h"
#include "global.h"
#include "token.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //
void lex_init ();
void lex_deinit ();
Token *lex_get (Token *oldToken);

// Internal //
bool mCheckForInclude ();
void mSkipBlankSpaces ();
Token *mGetWord ();
Token *mGetNumber ();
Token *mGetString ();
Token *mGetEOL ();
Token *mGetSpecial ();
Token *mGetEOF ();
Token *mGetError ();

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// lex_init
//// Initialize Lexer Module
void lex_init ()
{
  // Blank right now //
}

//// lex_deinit
//// Deinitialize Lexer Module
void lex_deinit ()
{
  // Blank right now //
}

//// lex_get
//// Get next token from buffer, delete the oldToken
Token *lex_get (Token *oldToken)
{
  char c;

  // This allows the current line to be incremented when this is
  //  the first token of a new line
  if (oldToken)
  {
    if (oldToken->code == ttEOL)
    {
      gnCurrentLine++;
    }
  }

  // Delete the old token //
  if (oldToken) tok_del (oldToken);

  mSkipBlankSpaces ();
  mCheckForInclude();
  mSkipBlankSpaces ();
  c = ib_get ();

  if (k_inSet (c, LEXER_LETTERS))       return mGetWord ();
  else if (k_inSet (c, LEXER_DIGITS))   return mGetNumber ();
  else if (c == '\"')                   return mGetString ();
  else if (c == 0)                      return mGetEOL ();
  else if (c == INPUTBUFFER_EOF)        return mGetEOF ();
  else                                  return mGetError ();
}

//// mCheckForInclude
//// Check for inlusion symbol '$'; return true if found
bool mCheckForInclude ()
{
  char *filename;
  int pos;

  if (ib_get() == '$')
  {
    filename = k_new (256);
    k_copy (filename, "");
    pos = 0;

    ib_seek ();
    while (ib_get() != 0)
    {
      filename[pos++] = ib_get();
      ib_seek ();
    }

    filename[pos - 1] = '\0';

    ib_include (filename);

    k_del (filename);

    return true;
  }

  return false;
}

//// mSkipBlankSpaces
//// Remove blank spaces from buffer; includes comments, and all whitespace
void mSkipBlankSpaces ()
{
  while (((ib_get() <= ' ') && (ib_get() > 0)) || (ib_get() == '#'))
  {
    if (ib_get() == '#')
    {
      while (ib_get() != 0)
      {
        ib_seek ();
      }
      if (ib_get() == 0)
      {
        // By returning now, lex_get() can return an EOL token //
        return;
      }
    }
    ib_seek ();
  }
}

//// mGetWord
//// Return a word from the buffer
Token *mGetWord ()
{
  // Vars //
  Token *token;
  char c;
  int nLexemePos;
  int nLen;

  token = tok_new();
  c = ib_get ();
  nLexemePos = 0;

  // Put the characters of the token into 'token->lexeme' //
  do {
    token->lexeme[nLexemePos++] = c;
    ib_seek();
    c = ib_get ();
  } while (k_inSet (c, LEXER_LETTERS) || k_inSet (c, LEXER_DIGITS));
  token->lexeme[nLexemePos] = '\0';

  // Length of the token //
  nLen = strlen (token->lexeme);

  // Assume it's an identifier //
  token->code = ttIdentifier;

  // Verify length of lexeme //
  if ((nLen >= LEXER_MIN_WORD_LEN) && (nLen <= LEXER_MAX_WORD_LEN))
  {
    switch (nLen)
    {
      case 2:
        if (strcmp (token->lexeme, "jt") == 0) token->code = ttJT;
        else if (strcmp (token->lexeme, "jf") == 0) token->code = ttJF;
        else if (strcmp (token->lexeme, "eq") == 0) token->code = ttEQ;
        else if (strcmp (token->lexeme, "lt") == 0) token->code = ttLT;
        else if (strcmp (token->lexeme, "gt") == 0) token->code = ttGT;
        else if (strcmp (token->lexeme, "or") == 0) token->code = ttOR;
        else if (strcmp (token->lexeme, "db") == 0) token->code = ttDB;
        else if (strcmp (token->lexeme, "di") == 0) token->code = ttDI;
        else if (strcmp (token->lexeme, "dr") == 0) token->code = ttDR;
        else if (strcmp (token->lexeme, "ds") == 0) token->code = ttDS;
        break;
      case 3:
        if (strcmp (token->lexeme, "nop") == 0) token->code = ttNOP;
        else if (strcmp (token->lexeme, "jmp") == 0) token->code = ttJMP;
        else if (strcmp (token->lexeme, "ret") == 0) token->code = ttRET;
        else if (strcmp (token->lexeme, "toi") == 0) token->code = ttTOI;
        else if (strcmp (token->lexeme, "tos") == 0) token->code = ttTOS;
        else if (strcmp (token->lexeme, "tor") == 0) token->code = ttTOR;
        else if (strcmp (token->lexeme, "tob") == 0) token->code = ttTOB;
        else if (strcmp (token->lexeme, "pop") == 0) token->code = ttPOP;
        else if (strcmp (token->lexeme, "neq") == 0) token->code = ttNEQ;
        else if (strcmp (token->lexeme, "and") == 0) token->code = ttAND;
        else if (strcmp (token->lexeme, "cat") == 0) token->code = ttCAT;
        else if (strcmp (token->lexeme, "cin") == 0) token->code = ttCIN;
        else if (strcmp (token->lexeme, "asc") == 0) token->code = ttASC;
        else if (strcmp (token->lexeme, "neg") == 0) token->code = ttNEG;
        else if (strcmp (token->lexeme, "not") == 0) token->code = ttNOT;
        break;
      case 4:
        if (strcmp (token->lexeme, "call") == 0) token->code = ttCALL;
        else if (strcmp (token->lexeme, "stop") == 0) token->code = ttSTOP;
        else if (strcmp (token->lexeme, "copy") == 0) token->code = ttCOPY;
        else if (strcmp (token->lexeme, "push") == 0) token->code = ttPUSH;
        else if (strcmp (token->lexeme, "lteq") == 0) token->code = ttLTEQ;
        else if (strcmp (token->lexeme, "gteq") == 0) token->code = ttGTEQ;
        else if (strcmp (token->lexeme, "cout") == 0) token->code = ttCOUT;
        else if (strcmp (token->lexeme, "rand") == 0) token->code = ttRAND;
        else if (strcmp (token->lexeme, "swap") == 0) token->code = ttSWAP;
        else if (strcmp (token->lexeme, "addi") == 0) token->code = ttADDI;
        else if (strcmp (token->lexeme, "subi") == 0) token->code = ttSUBI;
        else if (strcmp (token->lexeme, "muli") == 0) token->code = ttMULI;
        else if (strcmp (token->lexeme, "divi") == 0) token->code = ttDIVI;
        else if (strcmp (token->lexeme, "powi") == 0) token->code = ttPOWI;
        else if (strcmp (token->lexeme, "modi") == 0) token->code = ttMODI;
        else if (strcmp (token->lexeme, "addr") == 0) token->code = ttADDR;
        else if (strcmp (token->lexeme, "subr") == 0) token->code = ttSUBR;
        else if (strcmp (token->lexeme, "mulr") == 0) token->code = ttMULR;
        else if (strcmp (token->lexeme, "divr") == 0) token->code = ttDIVR;
        else if (strcmp (token->lexeme, "powr") == 0) token->code = ttPOWR;
        else if (strcmp (token->lexeme, "modr") == 0) token->code = ttMODR;
        else if (strcmp (token->lexeme, "popl") == 0) token->code = ttPOPL;
        else if (strcmp (token->lexeme, "popp") == 0) token->code = ttPOPP;
        else if (strcmp (token->lexeme, "true") == 0)
        {
          token->code = ttBoolLiteral;
          tok_set (token, k_new_b(true));
        }
        break;
      case 5:
        if (strcmp (token->lexeme, "pushl") == 0) token->code = ttPUSHL;
        else if (strcmp (token->lexeme, "pushp") == 0) token->code = ttPUSHP;
        else if (strcmp (token->lexeme, "dbgon") == 0) token->code = ttDBGON;
        else if (strcmp (token->lexeme, "fio_c") == 0) token->code = ttFIO_C;
        else if (strcmp (token->lexeme, "fio_r") == 0) token->code = ttFIO_R;
        else if (strcmp (token->lexeme, "fio_w") == 0) token->code = ttFIO_W;
        else if (strcmp (token->lexeme, "false") == 0)
        {
          token->code = ttBoolLiteral;
          tok_set (token, k_new_b(false));
        }
        break;
      case 6:
        if (strcmp (token->lexeme, "fio_or") == 0) token->code = ttFIO_OR;
        else if (strcmp (token->lexeme, "fio_ow") == 0) token->code = ttFIO_OW;
        break;
      case 7:
        if (strcmp (token->lexeme, "dbgline") == 0) token->code = ttDBGLINE;
        else if (strcmp (token->lexeme, "dbgfile") == 0) token->code = ttDBGFILE;
        else if (strcmp (token->lexeme, "fio_eof") == 0) token->code = ttFIO_EOF;
        break;
    }
  }

  return token;
}

//// mGetNumber
Token *mGetNumber ()
{
  // Variables //
  Token *token;
  int nPos;
  int nMantissaLen;

  token = tok_new();
  nPos = 0;
  nMantissaLen = -1; // Assume the token is an integer

  token->lexeme[nPos] = ib_get ();
  ib_seek ();

  while (k_inSet (ib_get(), LEXER_NUMBER_ELEMENTS))
  {
    if (nMantissaLen > -1) // if currently in mantissa, increase mantissa length
    {
      nMantissaLen++;
    }
    if (ib_get() == '.')
    {
      nMantissaLen++;
    }
    token->lexeme[++nPos] = ib_get();
    ib_seek ();
  }

  // Fill any empty mantissa places with zeros //
  if (nMantissaLen > -1)
  {
    while (nMantissaLen < GLOBAL_MANTISSA_SIZE)
    {
      nMantissaLen++;
      token->lexeme[++nPos] = '0';
    }
  }

  // End lexeme string //
  token->lexeme[++nPos] = 0;

  if (nMantissaLen > -1) // rational number
  {
    token->code = ttRatLiteral;
    tok_set (token, k_new_d(k_conv_sd(token->lexeme)));
  }
  else // integer number
  {
    token->code = ttIntLiteral;
    tok_set (token, k_new_n(k_conv_si(token->lexeme)));
  }

  return token;
}

//// mGetString
Token *mGetString ()
{
  // Variables //
  Token *token;
  char c;
  int pos;

  token = tok_new();
  pos = 0;

  // Token is a string literal //
  token->code = ttStrLiteral;

  // Opening quote //
  token->lexeme[pos++] = '\"';

  // Get the string //
  ib_seek (); // skip opening quote
  c = ib_get ();
  while (c != INPUTBUFFER_EOF)
  {
    if (c == '\"') // current char is quote
    {
      ib_seek ();
      c = ib_get ();
      if (c != '\"')
      {
        break; // not a double quote, so end the string
      }
    }
    else if (c == 0) // end of line reached
    {
      c = ' '; // replace end-line with blank
      err (true, EF_INCOMPLETE_STR_LIT, NULL);
      erre (EFF_IGNORE);
      break;
    }

    // Append current char to string //
    token->lexeme[pos++] = c;

    // Get next char //
    ib_seek ();
    c = ib_get ();
  }

  // BUG?:
  // Should possibly be removed, since the first char after a string may be eof
  if (c == INPUTBUFFER_EOF)
  {
    err (true, EF_UNEXPECTED_EOF, NULL);
    erre (EFF_IGNORE);
  }

  token->lexeme[pos++] = '\"'; // closing quote
  token->lexeme[pos] = '\0';

  // After this, token->lexeme will contain the string with quotes and //
  // token->value will contain the string minus the quotes             //
  tok_set (token, k_sub (token->lexeme, 1, strlen (token->lexeme) - 1));

  return token;
}

//// mGetEOL
Token *mGetEOL ()
{
  Token *token = tok_new();

  token->code = ttEOL;
  token->lexeme[0] = 0;

  ib_seek ();

  return token;
}

//// mGetEOF
Token *mGetEOF ()
{
  Token *token = tok_new();

  token->code = ttEOF;
  token->lexeme[0] = 0;

  return token;
}

//// mGetError
Token *mGetError ()
{
  Token *token = tok_new();

  token->code = ttError;
  token->lexeme[0] = ib_get ();
  token->lexeme[1] = '\0';

  ib_seek ();
  err (true, EF_UNDEFINED_TOKEN, token->lexeme);
  erre (EFF_IGNORE);

  return token;
}

#endif
